import re
from typing import Any as Any

class CollectionTranslator:
    """ Search strings, lists or dicts for placeholders using 
    a regular expression provided by the user. Replace the 
    matches with a value provided by a translator function, 
    which is as well provided by the user.
    For dicts and lists, the search is recursive.
    See constructor / functions for details.
    """

    def __init__(self, translator_func = None, regexp: str = r"(?<!\\){[^{}]*}"):
        """ Construct new CollectionTranslator object with default regular expression.

        Args:
            translator_func (_type_, optional): function to be called to
                translate a placeholder to its value. Defaults to None.
            regexp (str, optional): regular expression for finding
                placeholders. Defaults to r"(?<!\){[^{}]*}". That means
                that placeholders are marked by surrounding curly brackets,
                e.g. {KEY}.
        
        default regexp: a curly left bracket NOT preceded by a backslash starts
        a placeholder. It's name is denoted by the following sequence of characters
        NOT containing { or }. The end of the placeholder is a curly right bracket.

        translator_func and regexp may be changed later by assigning new
        values to self.translator_func and self.regexp
        """
        self.regexp = regexp;
        self.translator_func = translator_func;

    def translate(self, val: Any) -> Any:
        """ Convenience function which calls translate_dict, 
        translate_list or translate_str, depending on the type of val.
        For details: see these functions.

        Args:
            val (_Any_): value to be translated

        Returns:
            any: return value of the above functions
        """
        if isinstance(val, dict):
            # recursivley go down to nested dict
            val = self.translate_dict(val)
        elif isinstance(val, list):
            # recursivley go down to nested list
            val = self.translate_list(val)
        elif isinstance(val, str):
            # replace placeholders, if possible
            val = self.translate_str(val)
        #else: don't change
        return val
    
    # verwende eventuell str.format() für translate_str
    """
    { "schlüssel" : ["format string mit platzhaltern {0} und {1}", "key0", "key1"], ... }
    ->
    { "schlüssel" : "format string mit platzhaltern val0 und val1", ... }
    oder
    { "schlüssel" : { use_placeholders: true, 
                      format_string: "format string mit platzhaltern {0} und {1}",
                      keys: [ "key0", "key1" ]
                    }, ... }
    ->
    { "schlüssel" : "format string mit platzhaltern val0 und val1", ... }
    """
    def translate_str(self, s: str) -> str:
        r""" Search s for placeholders using self.regexp and
        replace them with a value provided by self.translator_func.
        Example: "{KEY}" -> "REPLACEMENT", if translator_func returns
        "REPLACEMENT" when calling translator_func("KEY").

        To insert a literal "{" or "}" into the string
        which is not to be treated as a bracket surrounding
        a placeholder, use "double-backslash-{" (or }) instead.
        To insert a literal backslash, use "double-backslash-/".
        Backslashes within placeholders are not allowed.

        Example when reading documentation:
        "val: \\\\{{KEY}\\\\} \\\\/x /y" results in
        "{REPLACEMENT} \\x /y"

        Same Example when reading source code:
        "val: \\{{KEY}\\} \\/x /y" results in
        "{REPLACEMENT} ", followed by backslash-x, "/y"

        Args:
            s (str): String with placeholders in curly brackets
            func (_type_): function to get replacement for placeholder

        Returns:
            str: String with replacements. s itself remains unchanged,
                 since strings are immutable.
        """
        if self.translator_func:
            # find and replace placeholders:
            mo = re.findall(self.regexp, s)
            for m in mo:
                val = self.translator_func(m[1:-1])
                s = re.sub(m, val, s, 1)
        # replace \\{, \\}, \\/ by {, }, \
        s = re.sub(r"\\{", "{", s)
        s = re.sub(r"\\}", "}", s)
        s = re.sub(r"\\/", r"\\", s)
        return s
    
    def translate_list(self, l: list) -> list:
        """ Translate all entries in the given list. Depending on the 
        type of the entry, self.translate_str, self.translate_list 
        or self.translate_dict is called.
        This combination of functions works recursively.
        
        The list items themselves are replaced by the translated versions.

        Args:
            l (list): List with items to be translated

        Returns:
            list: The (now changed) list itself.
        """
        # process all list items
        for i in range(len(l)):
            val = l[i]
            if isinstance(val, dict):
                # recursivley go down to nested dict
                self.translate_dict(l[i])
            elif isinstance(val, list):
                # recursivley go down to nested list
                self.translate_list(l[i])
            elif isinstance(val, str):
                # replace placeholders, if possible
                l[i] = self.translate_str(val)
            #else: ignore entry
        return l

    def translate_dict(self, d: dict) -> dict:
        """ Translate all values in the given dict. Depending on the 
        type of the value, self.translate_str, self.translate_list 
        or self.translate_dict is called. The keys are left unchanged.
        This combination of functions works recursively.
        
        The dict values themselves are replaced by the translated versions.

        Args:
            d (dict): dict with items to be translated

        Returns:
            list: The (now changed) dict itself.
        """
        # process whole dict
        for key in d:
            val = d[key]
            if isinstance(val, dict):
                # recursivley go down to nested dict
                self.translate_dict(val)
            elif isinstance(val, list):
                # recursivley go down to nested list
                self.translate_list(val)
            elif isinstance(val, str):
                # replace placeholders, if possible
                d[key] = self.translate_str(val)
            #else: ignore entry
        return d



"""
# collectiontools_vrb

Various tools to handle collection I/O and transformations.

  - On PyPI: https://pypi.org/project/collectiontools-vrb/
  - On GitHub: https://github.com/v-r-b/collectiontools_vrb 

## \_\_init__.py

```function update_dict_from_url()```

Read collection data from the given URL. 

```function update_dict_from_key_value_file()```

Read collection data from the given file. 

```function dict_contains_path()```

Tests if a given path is a valid key inside the given dict.

```function update_dict_from_json_file()```
  
Read collection data from config file. Check for mandatory keys.

## collectiontranslator.py

```class CollectionTranslator```

Search strings, lists or dicts for placeholders using a 
regular expression provided by the user. Replace the matches 
with a value provided by a translator function, which is 
as well provided by the user.
"""

from __future__ import annotations
    
import asyncio
from types import FrameType
import aiohttp
import json
import traceback
import sys, inspect
from typing import Any as Any

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from _typeshed import SupportsWrite

from collectiontools_vrb.collectiontranslator import *

async def _update_dict_from_url(d: dict, url: str, sep: str, strip: bool) -> aiohttp.ClientResponse:
    """Internal method using asyncio to carry out
    the task of updateFromURL.
    See: updateDictFromURL(d, url, sep)
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            if response.ok:
                text: str = await response.text()
            else:
                raise Exception(
                    f"Unable to read from URL {url}. Status {response.status}")
            for l in text.splitlines():
                parts = l.split(sep, 1)
                if len(parts) == 2:
                    if strip:
                        d[parts[0].strip()] = parts[1].strip()
                    else:
                        d[parts[0]] = parts[1]
    return response
    
def update_dict_from_url(d: dict, url: str, *,
            sep: str = "=", strip: bool = True,
            reraise_exc: bool = False, 
            print_errors_to: SupportsWrite[str]|None = None) -> bool:
    """Read collection data from the given URL. 

    The data must consist of lines of key-value-pairs
    separated by the given sep string. Sample line:
    If sep == "=", NAME=Miller results to a key-value-pair "NAME":"Miller".

    Args:
        d (dict): dictionary to be updated
        url (str): URL to read from
        sep (str, optional): String to separate key from value. Defaults to "=".
        strip (bool, optional): Strip trailing and leading whitespace from keys and values? Defaults to True.
        reraise_exc (bool, optional): reraise caught exceptions if True, otherwise
            print exception info if print_errors == True. Defaults to False.
        print_errors_to (SupportsWrite[str], optional): stream to print error information to, 
            i.e. missing mandatory keys, error on opening file etc. Defaults to None.
            If not None, print_errors_to will be passed to the print function as "file=" argument.

    Raises:
        Exception: if URL can not be read properly (only if reraise_exc == True)

    Returns:
        True, if HTTP status code of get(url) operation is less than 400, False otherwise
    """
    try:
        response = asyncio.run(_update_dict_from_url(d, url, sep, strip))
    except Exception as exc:
        # reraise exception if raiseExc == True
        if reraise_exc:
            raise
        # else print error if requested to do so
        elif print_errors_to:
            print(traceback.format_exc(), file=print_errors_to)
        return False
    return response.ok

def update_dict_from_key_value_file(d: dict, path: str, *,
            sep: str = "=", strip: bool = True,
            reraise_exc: bool = False, 
            print_errors_to: SupportsWrite[str]|None = None) -> bool:
    """Read collection data from the given file. 

    The data must consist of lines of key-value-pairs
    separated by the given sep string. Sample line:
    If sep == "=", NAME=Miller results to a key-value-pair "NAME":"Miller".

    Reraise an exception (e.g. FileNotFound) only if argument == True, 
    otherwise print exception information only.

    Args:
        d (dict): dictionary to be updated
        path (str): file descriptor or path to open
        sep (str, optional): String to separate key from value. Defaults to "=".
        strip (bool, optional): Strip trailing and leading whitespace from keys and values? Defaults to True.
        reraise_exc (bool, optional): reraise caught exceptions if True, otherwise
            print exception info if print_errors == True. Defaults to False.
        print_errors_to (SupportsWrite[str], optional): stream to print error information to, 
            i.e. missing mandatory keys, error on opening file etc. Defaults to None.
            If not None, print_errors_to will be passed to the print function as "file=" argument.

    Returns:
        bool: True on success, False otherwise

    Raises:
        Exception: FileNotFound (only if reraise_exc == True)
    """
    try:
        with open(path) as f:
            text: list[str] = f.read().splitlines()
        for l in text:
            parts = l.split(sep, 1)
            if len(parts) == 2:
                if strip:
                    d[parts[0].strip()] = parts[1].strip()
                else:
                    d[parts[0]] = parts[1]
        return True
    except Exception as exc:
        # reraise exception if raiseExc == True
        if reraise_exc:
            raise
        # else print error if requested to do so
        elif print_errors_to:
            print(traceback.format_exc(), file=print_errors_to)
        return False

def dict_contains_path(nested_dict: dict, keypath: str, *, sep: str = ".") -> bool:
    """ Tests if a given path is a valid key inside the given dict.
    It is assumed, that the key path elements are seperated by ".".
    To change this, use argument sep=...

    Args:
        nested_dict (dict): (nested) dictionary to search in
        keypath (str): key to be found (elements separated by sep)
        sep (str, optional): key element separator to be used. Defaults to ".".

    Returns:
        bool: True if the key could be found, False otherwise
    """
    # Split key path in first elements and all the rest
    key_parts = keypath.split(sep, 1)
    num_key_parts = len(key_parts)
    if num_key_parts == 1:
        # found single key. Test if it is in the dict:
        return key_parts[0] in nested_dict
    else:
        # test for first key element. 
        if key_parts[0] in nested_dict:
            # if found, go one level deeper with the nested dict
            # found at [key_parts[0]] und the rest of the key
            dict_part = nested_dict[key_parts[0]]
            return dict_contains_path(dict_part, key_parts[1], sep=sep)
        else:
            # if not found, there's no need to process
            # the rest of the key
            return False

def update_dict_from_json_file(d: dict, file: str, *,
                               mandatory_keys: list[str] = [], 
                               reraise_exc: bool = False,
                               print_errors_to: SupportsWrite[str]|None = None) -> bool:
    """ Read collection data from config file. Check for mandatory keys.
    Reraise an exception (e.g. FileNotFound) only if reraise_exc == True, 
    otherwise print exception information only if print_errors_to is not None.

    Args:
        d (dict): dictionary to be updated
        file (str): file descriptor or path to open
        mandatory_keys (list[str], optional): check for these keys in the dict
            If not found, return false and don't update dict. Defaults to [].
        reraise_exc (bool, optional): reraise caught exceptions if True, otherwise
            print exception info if print_errors == True. Defaults to False.
        print_errors_to (SupportsWrite[str], optional): stream to print error information to, 
            i.e. missing mandatory keys, error on opening file etc. Defaults to None.
            If not None, print_errors_to will be passed to the print function as "file=" argument.

    Returns:
        bool: True on success, False otherwise
    """

    result = None
    try:
        # read config file:
        with open(file) as fp:
            result = json.load(fp)

        # check for mandatory keys:
        for key in mandatory_keys:
            if not dict_contains_path(result, key, sep="."):
                if print_errors_to:
                    cf = inspect.currentframe()
                    if isinstance(cf, FrameType):
                        print(cf.f_code.co_name,
                              f": missing key '{key}' in file '{file}'", 
                              file=print_errors_to)
                return False

        # update d only in case all mandatory keys could be found
        d.update(result)
        return True
    except Exception as exc:
        # reraise exception if raiseExc == True
        if reraise_exc:
            raise
        # else print error if requested to do so
        elif print_errors_to:
            print(traceback.format_exc(), file=print_errors_to)
        return False
